import datetime
import json
from os import path
import typing
import urllib.parse

import bleach
import bs4
import horkos
import jinja2
import markdown
import yaml

from horkos_doc import _definitions
from horkos_doc import _site_generator

PACKAGE_DIR = path.dirname(path.realpath(__file__))
RESOURCE_DIR = path.join(PACKAGE_DIR, 'resources')


def first_paragraph(value: str) -> str:
    """Extract the first paragraph from a markdown string."""
    rendered = render_markdown(value)
    soup = bs4.BeautifulSoup(rendered, features="html.parser")
    return str(soup.find('p'))


def shorten(value: typing.Any) -> str:
    """
    Attempt to render the given value, shorten it if it is more
    than 20 characters.
    """
    result = json.dumps(value)
    if len(result) < 20:
        return result
    return f'{result[:16]}...'


def render_markdown(value: str) -> str:
    """Render the given markdown."""
    result = markdown.markdown(value, extensions=['fenced_code'])
    allowed = [
        'a', 'abbr', 'acronym', 'b', 'blockquote', 'code', 'em', 'i', 'li',
        'ol', 'strong', 'ul', 'p', 'pre'
    ]
    return bleach.clean(result, tags=allowed)


def yamilify_field(field: horkos.Field) -> str:
    """Convert a field to yaml."""
    return yaml.dump({
        'required': field.required,
        'nullable': field.nullable,
        'checks': [
            c.name if not c.args else {'name': c.name, 'args': c.args}
            for c in field.checks
        ],
        'labels': field.labels,
    }, sort_keys=False)


def yamilify(value: typing.Any) -> str:
    """Convert anything to yaml."""
    return yaml.dump(value, sort_keys=False)


def _add_schema_files(
        gen: _site_generator.SiteGenerator,
        catalog: horkos.Catalog,
        context: _definitions.Context,
):
    schemas = sorted(catalog.schemas, key=lambda s: s.name)
    for i, schema in enumerate(schemas):
        bottom = max(i - 15, 0)
        top = bottom + 30
        if top > len(schemas):
            top = len(schemas)
            bottom = max(top - 30, 0)
        neighbors = schemas[bottom:top]
        cleaned = schema.name.replace(' ', '-').replace('/', '-')
        page_path = f'schemas/{cleaned}.html'
        gen.add_page(
            f'schemas/{schema.name}',
            page_path,
            'dataset.html',
            {
                'schema': schema,
                'context': context,
                'links': [
                    {
                        'to': f'schemas/{s.name}',
                        'text': s.name,
                        'bold': s.name == schema.name,
                    }
                    for s in neighbors
                ],
            },
        )

def _add_index_list(
        gen: _site_generator.SiteGenerator,
        catalog: horkos.Catalog,
        context: _definitions.Context,
):
    """Render a listing of all schemas in the catalog."""
    schemas = sorted(catalog.schemas, key=lambda s: s.name)
    gen.add_page(
        'index',
        'index.html',
        'listing.html',
        {'schemas': schemas, 'context': context}
    )


def _add_static_files(gen: _site_generator.SiteGenerator):
    gen.add_page('static/style', 'static/style.css', 'style.css')
    gen.add_page(
        'static/expand_script', 'static/expand_script.js', 'expand_script.js'
    )


def _add_search(
        gen: _site_generator.SiteGenerator,
        catalog: horkos.Catalog,
        context: _definitions.Context,
):
    gen.add_page('search', 'search.html', 'search.html')
    results = []
    for schema in catalog.schemas:
        base_url = gen.relative_url('search', f'schemas/{schema.name}')
        results.append({
            'title': [schema.name],
            'text': schema.description,
            'html': first_paragraph(schema.description),
            'link': base_url,
        })
        for field in schema.fields:
            results.append({
                'title': [schema.name, field.name],
                'text': field.description,
                'html': first_paragraph(field.description),
                'link': f'{base_url}#{urllib.parse.quote(field.name)}',
            })
    size = 100
    chunks = [results[i:i+size] for i in range(0, len(results), size)]
    search_indices = []
    for i, chunk in enumerate(chunks):
        gen.add_page(
            f'search-resources/{i}',
            f'search-resources/{i}.json',
            'raw_json.json',
            {'body': json.dumps(chunk)},
        )
        search_indices.append(
            gen.relative_url('search', f'search-resources/{i}')
        )
    # Add it again to use new args
    gen.add_page(
        'search', 'search.html', 'search.html',
        {'indices': json.dumps(search_indices), 'context': context}
    )


def render_catalog(
        catalog: horkos.Catalog,
        context: _definitions.Context = None,
):
    """Render the given catalog into the dest_path directory."""
    context = context or _definitions.Context()
    env = jinja2.Environment(
        loader=jinja2.PackageLoader('horkos_doc', 'templates'),
        autoescape=jinja2.select_autoescape(['html', 'xml'])
    )
    env.filters['shorten'] = shorten
    env.filters['markdown'] = render_markdown
    env.filters['field_to_yaml'] = yamilify_field
    env.filters['to_yaml'] = yamilify
    env.filters['first_paragraph'] = first_paragraph
    env.globals['now'] = datetime.datetime.utcnow
    gen = _site_generator.SiteGenerator(env)
    _add_schema_files(gen, catalog, context)
    _add_static_files(gen)
    _add_index_list(gen, catalog, context)
    _add_search(gen, catalog, context)
    gen.render(context.build_dir)



import os
from os import path
import tempfile

import horkos
import horkos_doc
from horkos_doc import _renderer

MY_DIR = path.dirname(__file__)
RESOURCE_DIR = path.join(MY_DIR, 'resources')
SCHEMAS_DIR = path.join(RESOURCE_DIR, 'schemas')


def create_test_catalog() -> horkos.Catalog:
    """Create a generic catalog."""
    return horkos.Catalog(
        path.join(SCHEMAS_DIR, f) for f in os.listdir(SCHEMAS_DIR)
    )


def test_render_catalog_without_error():
    """Should be able to render a catalog without error"""
    catalog = create_test_catalog()
    with tempfile.TemporaryDirectory() as tmpdirname:
        context = horkos_doc.Context(build_dir=tmpdirname)
        _renderer.render_catalog(catalog, context)



import os
from os import path
import urllib.parse

import jinja2


class SiteGenerator:
    """class for aggregating site pages and rendering them to html."""

    def __init__(self, environment: jinja2.Environment):
        self._pages = {}
        self.environment = environment
        self.environment.filters['relative_url'] = self._get_path_for

    @jinja2.contextfilter
    def _get_path_for(
            self, context: jinja2.runtime.Context, page_name: str
    ) -> str:
        """
        Get the relative path from the page being rendered within the context
        to the given page_name.
        """
        return self.relative_url(context['_rendering_page_name'], page_name)


    def relative_url(self, source_page: str, dest_page: str) -> str:
        source_page = self._pages[source_page]
        dest_page = self._pages[dest_page]
        source_path = list(source_page['page_path'].split('/'))[:-1]
        dest_path = list(dest_page['page_path'].split('/'))

        diverge_at = 0
        for i, pair in enumerate(zip(source_path, dest_path)):
            diverge_at = i
            if pair[0] != pair[1]:
                break
        escaped = [urllib.parse.quote(s) for s in dest_path[diverge_at:]]
        back_up = len(source_path) - diverge_at
        return '/'.join(['..'] * back_up + escaped)


    def add_page(
            self,
            page_name: str,
            page_path: str,
            template_name: str,
            template_args: dict = None,
    ):
        args = dict(template_args or {})
        args['_rendering_page_name'] = page_name
        self._pages[page_name] = {
            'page_name': page_name,
            'page_path': page_path,
            'template_name': template_name,
            'template_args': args
        }

    def render(self, build_dir: str):
        for page in self._pages.values():
            template = self.environment.get_template(page['template_name'])
            rendered = template.render(
                **page['template_args'],
            )
            write_path = path.join(build_dir, page['page_path'])
            dirname = path.dirname(write_path)
            if not path.exists(dirname):
                os.makedirs(dirname)
            with open(write_path, 'w') as f:
                f.write(rendered)



import os
from os import path
import tempfile

import horkos
import pytest

import horkos_doc
from horkos_doc import _cmdline

MY_DIR = path.dirname(__file__)
RESOURCE_DIR = path.join(MY_DIR, 'resources')
SCHEMAS_DIR = path.join(RESOURCE_DIR, 'schemas')
CONFIG_FILE = path.join(RESOURCE_DIR, 'horkos-doc.config')


def test_render_catalog_happy_path():
    """Should be able to generate html from yaml files."""
    with tempfile.TemporaryDirectory() as tmpdirname:
        _cmdline.main(['-c', CONFIG_FILE, '-o', tmpdirname, SCHEMAS_DIR])

        generated_root = os.listdir(tmpdirname)
        schemas_html = os.listdir(path.join(tmpdirname, 'schemas'))

    assert 'index.html' in generated_root
    assert len(schemas_html) > 2


def test_render_catalog_without_config():
    """Should be able to generate html from yaml files without config."""
    with tempfile.TemporaryDirectory() as tmpdirname:
        _cmdline.main(['-o', tmpdirname, SCHEMAS_DIR])

        generated_root = os.listdir(tmpdirname)
        schemas_html = os.listdir(path.join(tmpdirname, 'schemas'))

    assert 'index.html' in generated_root
    assert len(schemas_html) > 2


def test_render_catalog_fails_without_input_arg():
    """Should exit if `input` is omitted."""
    with pytest.raises(SystemExit):
        with tempfile.TemporaryDirectory() as tmpdirname:
            _cmdline.main(['-o', tmpdirname])


def test_render_catalog_fails_without_valid_schema():
    """Should exit if a parsed schema is bad."""
    schema_dir = path.join(RESOURCE_DIR, 'erant-schemas')
    with pytest.raises(SystemExit):
        with tempfile.TemporaryDirectory() as tmpdirname:
            _cmdline.main(['-o', tmpdirname, schema_dir])



import glob
from os import path
import collections
import setuptools

MY_DIR = path.abspath(path.dirname(__file__))
with open(path.join(MY_DIR, 'README.md'), encoding='utf-8') as f:
    LONG_DESCRIPTION = f.read()

setuptools.setup(
    name='horkos-doc',
    version='0.0.3',
    description=(
        'A library for building data documentation.'
    ),
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    author='Kevin Schiroo',
    author_email='kjschiroo@gmail.com',
    license='MIT',
    url='https://gitlab.com/kjschiroo/horkos-doc',
    entry_points = {
        'console_scripts': ['horkos-doc=horkos_doc._cmdline:main'],
    },
    packages=setuptools.find_packages(),
    install_requires=[
        'beautifulsoup4', 'bleach', 'jinja2', 'markdown', 'horkos'
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
    ],
    python_requires='>=3.7',
)



import argparse
import configparser
import os
from os import path
import re
import sys

import horkos
import horkos.errors

from horkos_doc import _definitions
from horkos_doc import _renderer

SCHEMA_FILE_RE = re.compile(r'.*\.(yaml|yml|json)$')


def _parse_config_file(config_file: str) -> dict:
    """Parse an incoming config file and extract all top level keys."""
    if not path.isfile(config_file):
        return {}
    parser = configparser.ConfigParser()
    parser.read(config_file)
    args = {
        'author': parser['default'].get('author'),
        'home_page': parser['default'].get('home_page'),
        'output': parser['default'].get('output'),
        'input': parser['default'].get('input'),
    }
    return {k: v for k, v in args.items() if v is not None}


def _parse(args: list = None) -> dict:
    """
    Parse the incoming command-line arguments as well as configuration file
    based arguments to come up with a working set of args.
    """
    parser = argparse.ArgumentParser('horkos-doc')
    defaults = {}
    parser.add_argument(
        '--config', '-c',
        help=(
            'The path to a configuration file for horkos-doc. This file can '
            'be used to override the built-in default for command-line '
            'arguments. By default this is `.horkos_doc`.'
        )
    )
    defaults['config'] = '.horkos_doc'
    parser.add_argument(
        '--author', '-a',
        help=(
            'The author of the documentation. This is used to create '
            'copyright statements within the generated documentation.'
        )
    )
    defaults['author'] = None
    parser.add_argument(
        '--home-page', '--hp',
        help=(
            'The name to use for the home page of the generated documentation.'
            ' By default this is "Horkos Doc".'
        ),
    )
    defaults['home_page'] = 'Horkos Doc'
    output_dir = path.join('.', 'build')
    parser.add_argument(
        '--output', '-o',
        help=(
            'The directory in which to store built html. By default this is '
            f'{output_dir}'
        ),
    )
    defaults['output'] = output_dir
    parser.add_argument(
        'input',
        nargs='?',
        help=(
            'A directory of schema files from which to build the html '
            'documentation. This is not required if a value is set within '
            'the given config file.'
        )
    )
    parsed = parser.parse_args(args)
    config_parsed = _parse_config_file(parsed.config or defaults['config'])

    result = defaults.copy()
    result.update({k: v for k, v in config_parsed.items() if v is not None})
    result.update({k: v for k, v in vars(parsed).items() if v is not None})
    if result.get('input') is None:
        parser.print_help(sys.stderr)
        msg = (
            '\nERROR: `input` must be specified on the command-line if '
            'not specified in the config file.'
        )
        print(msg, file=sys.stderr)
        sys.exit(1)
    return result


def _get_schema_file_paths(input_dir: str) -> list:
    """Get all of the schema files within an input directory."""
    result = []
    for root, _, files in os.walk(input_dir):
        result.extend(
            path.abspath(path.join(root, f))
            for f in files
            if SCHEMA_FILE_RE.match(f)
        )
    return result


def main(args: list = None):
    """Convert a set of horkos schema files into a static website."""
    parsed = _parse(args)
    schema_files = _get_schema_file_paths(parsed['input'])
    catalog = horkos.Catalog()
    for schema_file in schema_files:
        try:
            catalog.update(schema_file)
        except horkos.errors.SchemaValidationError:
            print(f'{schema_file} is not a valid schema file')
            sys.exit(1)
    context = _definitions.Context(
        build_dir=parsed['output'],
        home_page=parsed['home_page'],
        author=parsed['author'],
    )
    _renderer.render_catalog(catalog, context)



import typing


class Context(typing.NamedTuple):
    """The configuration to use while rendering the catalog."""

    build_dir: str = 'build'
    home_page: str = 'Horkos Docs'
    author: str = None
